package com.jdbernard.teammaker

import groovy.beans.Bindable
import groovy.swing.SwingBuilder
import java.awt.GridBagConstraints as GBC
import java.awt.GridBagLayout
import java.awt.BorderLayout
import java.awt.event.MouseEvent
import javax.swing.DefaultListModel
import javax.swing.JFrame
import javax.swing.JOptionPane
import org.apache.log4j.Logger

public class TeamMaker {

    /* ======== MODEL ======== */

    public static def swing = new SwingBuilder()
    public static final String version = "0.4"

    // GUI Components
    def frame               // main frame
    def team1List           // JList for team 1
    def team2List           // JList for team 2
    def team1WinsButton 
    def team2WinsButton
    def newGameButton
    def sittingList         // JList for sitting players

    def inGamePlayerDropMenu
    def sittingPlayerDropMenu

    protected Logger log = Logger.getLogger(getClass())

    PlayerChooser playerChooser = new WeightedChooser(hardLimit: 2)
    OddsCalculator oddsCalculator = new OddsCalculator()

    // Data components
    class Observables {
        @Bindable List team1Players = []
        @Bindable List team2Players = []
        @Bindable List sittingPlayers = []
        @Bindable boolean inGame = false
        @Bindable int teamSize = 4
        @Bindable List popupPlayers = null
        @Bindable int popupTeam;
    }

    def model = new Observables()


    /* ======== VIEW ======== */

        /*
           +--------------------------------------------------+
           | JDB TeamMaker v0.2                          _[]X |
           +--------------------------------------------------+
           |                                                  |
           | +-Current-Game--------------+ +-Sitting-Players+ |
           | |                           | |                | |
           | | Team A          Team B    | | Jeptha       3 | |
           | |   Derrick         Eric    | | Jack         2 | |
           | |   Shane           Brandon | | Sam          2 | |
           | |   Shane           Darren  | | Greg         2 | |
           | |   Christian       Brett   | | Frank        1 | |
           | +---------------------------+ | Chris D      1 | |
           |                               | Joshua       1 | |
           |                               |                | |
           |                               |                | |
           |                               |                | |
           |                               |                | |
           |                               |                | |
           |                               +----------------+ |
           +--------------------------------------------------+

         */

    private void init() {
        
        inGamePlayerDropMenu = swing.popupMenu(/*'Player Options'*/) {
            menuItem('Bench Player',/* enabled: bind { !model.inGame },*/
                actionPerformed: { benchPlayer(model.popupPlayers,
                    model.popupTeam) })
        }

        sittingPlayerDropMenu = swing.popupMenu(/*'Player Options'*/) {
            menuItem('Assign to Team A', /*enabled: bind { !model.inGame },*/
                actionPerformed: {
                    model.popupPlayers.each { assignPlayer(it, 1) }})

            menuItem('Assign to Team B', /*enabled: bind { !model.inGame },*/
                actionPerformed: {
                    model.popupPlayers.each { assignPlayer(it, 2) }})

            menuItem('Set Games Sat',
                actionPerformed: {
                    model.popupPlayers.each { setGamesSat(it) }})
        }

        frame = swing.frame(title: "JDB TeamMaker v$version",
            size: [600, 400], locationRelativeTo: null,
            defaultCloseOperation: JFrame.EXIT_ON_CLOSE) {

            panel() {
                gridBagLayout()

                panel(constraints: gbc(gridx: 0, gridy: 0, fill: GBC.BOTH,
                        insets: [5, 5, 5, 0], weightx: 3, weighty: 1),
                    border: titledBorder(title: 'Current Game')) {
                    
                    gridBagLayout()

                    label('Team A', constraints: gbc(gridx: 0, gridy: 0,
                        fill: GBC.BOTH, insets: [5, 5, 0, 0]))

                    label('Team B', constraints: gbc(gridx: 1, gridy: 0, 
                        fill: GBC.BOTH, insets: [5, 5, 0, 5]))
                    
                    def teamListRenderer = new PlayerListCellRenderer(this,
                        showStats: false, colored: false)

                    team1List = list(cellRenderer: teamListRenderer,
                        constraints: gbc(gridx: 0, gridy: 1, fill: GBC.BOTH,
                        insets: [5, 5, 0, 0], weightx: 2, weighty: 2),
                        model: new DefaultListModel(),
                        mouseClicked: { evt ->
                            if (evt.button != MouseEvent.BUTTON3) return
                            showInGamePlayerDropMenu(evt)
                        })

                    team1WinsButton = button('Team A Wins',
                        constraints: gbc(gridx: 0, gridy: 2, anchor:
                            GBC.CENTER, insets: [5, 5, 5, 0]),
                        enabled: bind { model.inGame },
                        actionPerformed: { declareLoser(2)})

                    team2List = list(cellRenderer: teamListRenderer,
                        constraints: gbc(gridx: 1, gridy: 1, fill: GBC.BOTH,
                        insets: [5, 5, 0, 5], weightx: 2, weighty: 2),
                        model: new DefaultListModel(),
                        mouseClicked: { evt ->
                            if (evt.button != MouseEvent.BUTTON3) return
                            showInGamePlayerDropMenu(evt)
                        })

                    team2WinsButton = button('Team B Wins',
                        constraints: gbc(gridx: 1, gridy: 2,
                            anchor: GBC.CENTER, insets: [5, 5, 5, 5]),
                        enabled: bind { model.inGame },
                        actionPerformed: { declareLoser(1)} )

                }

                scrollPane(constraints: gbc(gridx: 1, gridy: 0, gridwidth: 2,
                        fill: GBC.BOTH, insets: [5, 5, 0, 5], weighty: 1),
                    border: titledBorder(title: 'Sitting Players')) {

                    sittingList = list(cellRenderer: new PlayerListCellRenderer(
                            this, showStats: true, colored: true),
                            model: new SortedPlayerModel(),
                            mouseClicked: { evt ->
                                if (evt.button != MouseEvent.BUTTON3) return
                                showSittingPlayerDropMenu(evt)
                            })
                }

                button('Next Game', constraints: gbc(gridx: 0, gridy: 1,
                    anchor: GBC.CENTER, insets: [5, 5, 5, 0]),
                    enabled: bind { !model.inGame},
                    actionPerformed: { newGame() })

                button('Add Player', constraints: gbc(gridx: 1, gridy: 1,
                    anchor: GBC.CENTER, insets: [5, 5, 5, 0]),
                    actionPerformed: { addPlayer() })

                newGameButton = button('Delete Player',
                    constraints: gbc(gridx: 2, gridy: 1, anchor: GBC.CENTER,
                        insets: [5, 5, 5, 5]),
                    actionPerformed: {
                        model.sittingPlayers.remove(sittingList.selectedValues)
                            refreshGUI() })

            }
        }

    }


    

    /* ======== CONTROLLER ======== */

    public static void main(String[] args) {
        def inst = new TeamMaker()
        inst.frame.show()
    }

    TeamMaker() {
        init()
    }

    /**
     * @return the number of players needed to round out even teams.
     */
    public int getSpotsOpen() {
        def spots =  (model.teamSize - model.team1Players.size()) +
                     (model.teamSize - model.team2Players.size())
        return (spots == 0 ? model.teamSize : spots)
    }

    private void addPlayer() {
        def name = JOptionPane.showInputDialog(frame,
            "Enter the new player's name: ", "New Player...",
            JOptionPane.QUESTION_MESSAGE)

        if (!name) return

        def player = new Player()
        player.name = name
        player.gamesSat = 0
        model.sittingPlayers << player

        refreshGUI()
    }

    private void newGame() {
        if (model.team1Players.size() < model.teamSize)
            populate(1)

        if (model.team2Players.size() < model.teamSize)
            populate(2)

        recalculateOdds()

        model.inGame = true

    }

    private void declareLoser(int teamNum) {
        model.sittingPlayers.each { it.gamesSat++ }
        def teamPlayers = playersForTeam(teamNum)
        teamPlayers.each { player ->
            player.gamesSat = 0
            model.sittingPlayers << player
        }

        teamPlayers.clear()

        refreshGUI()

        model.inGame = false
    }

    private void refreshGUI() {

        if (log.isTraceEnabled()) log.trace("Refreshing GUI.")
        if (log.isDebugEnabled()) {
            log.debug("team1Players: ${model.team1Players}")
            log.debug("team2Players: ${model.team2Players}")
        }

        team1List.model = new DefaultListModel()
        team2List.model = new DefaultListModel()
        sittingList.model = new SortedPlayerModel()

        model.team1Players.each { team1List.model.addElement(it) }
        model.team2Players.each { team2List.model.addElement(it) }
        model.sittingPlayers.each { sittingList.model.addElement(it) }

        frame.repaint()

        recalculateOdds()

    }

    private void recalculateOdds() {
        def players = model.sittingPlayers
        swing.doOutside {
            oddsCalculator.recalculate(players, getSpotsOpen(), playerChooser)
            swing.edt { sittingList.repaint() }
        }

    }

    private void populate(int teamNum) {
        def teamPlayers = playersForTeam(teamNum)
        while (teamPlayers.size() < model.teamSize) {
            def player = playerChooser.choose(model.sittingPlayers)
            teamPlayers << player
            model.sittingPlayers.remove(player)
        }

        refreshGUI();
    }

    private benchPlayer(Player p, int teamNum) {
        def teamPlayers = playersForTeam(teamNum)
        teamPlayers.remove(p)
        model.sittingPlayers << p
        refreshGUI()
    }

    private assignPlayer(Player p, int teamNum) {
        def teamPlayers = playersForTeam(teamNum)
        model.sittingPlayers.remove(p)
        teamPlayers << p
        refreshGUI()
    }

    private void showSittingPlayerDropMenu(def evt) {
        int index = sittingList.locationToIndex(evt.point)
        if (sittingList.isSelectedIndex(index))
            this.model.popupPlayers = sittingList.selectedValues
        else {
            sittingList.selectedIndex = index
            this.model.popupPlayers = [sittingList.model.getElementAt(index)]
        }

        sittingPlayerDropMenu.show(sittingList, evt.point.@x, evt.point.@y)
    }

    private void showInGamePlayerDropMenu(def evt) {
        def teamList = evt.source
        int index = teamList.locationToIndex(evt.point)
        if (teamList.isSelectedIndex(index)) 
            this.model.popupPlayers = teamList.selectedValues 
        else {
            teamList.selectedIndex = index
            model.popupPlayers = [teamList.model.getElementAt(index)]
        }

        this.model.popupTeam = (teamList == team1List ? 1 : 2)

        inGamePlayerDropMenu.show(teamList, evt.point.@x, evt.point.@y)
    }

    private void setGamesSat(Player p) {
        String value = JOptionPane.showInputDialog(frame,
            "How many games has ${p.name} sat?", "Set Games Sat.",
            JOptionPane.QUESTION_MESSAGE)
        int newGames

        if (!value) return
        try { newGames = value.toInteger() }
        catch (Throwable t) {
            JOptionPane.showMessageDialog(frame,
                "The value $value is not a valid number.",
                "Invalid Input", JOptionPane.ERROR_MESSAGE)
            setGamesSat(p)
            return
        }
        
        p.gamesSat = newGames

        refreshGUI()
    }

    private List playersForTeam(int teamNum) {
        if (teamNum == 1) return model.team1Players
        else return model.team2Players
    }

    private List listForTeam(int teamNum) {
        if (teamNum == 1) return team1List
        else return team2List
    }
}
